(* ProofPre includes proofs specific to
   constructs defined in the Coq Standard Library. *)



Require Import Le.
Require Import Lt.
Require Import Gt.
Require Import Plus.
Require Import List.
Require Import SemanticsBase.



(* ~~~~~~~ PROOFS ABOUT IN ~~~~~~~ *)

Definition not_in_neq {p1 p2 : Probe} (ne : p1 <> p2) : ~ In p1 (p2 :: nil) :=
  fun i => match in_inv i with
             | or_introl e => match ne (eq_sym e) with end
             | or_intror i' => match in_nil i' with end
           end.

Theorem in_remove_neq : forall {p1 p2 l} (ne : p1 <> p2) (i : In p1 l), In p1 (remove eq_Probe_dec p2 l).
intros. induction l; simpl.
  inversion i.
  inversion i; clear i.
    subst. destruct (eq_Probe_dec p2 p1).
      destruct (ne (eq_sym e)).
      simpl. left. reflexivity.
    apply IHl in H. clear IHl. destruct (eq_Probe_dec p2 a).
      assumption.
      right. assumption.
Qed.

Theorem in_remove_weak : forall {p1 p2 l}, In p1 (remove eq_Probe_dec p2 l) -> In p1 l.
intros. induction l; simpl in *.
  inversion H.
  destruct (eq_Probe_dec p2 a).
    subst. right. apply IHl. assumption.
    simpl in H. destruct H.
      left. assumption.
      right. apply IHl. assumption.
Qed.



(* ~~~~~~~ PROOFS ABOUT COUNT_OCC ~~~~~~~ *)

Definition count_occ_not_In {l p} : ~ In p l <-> count_occ eq_Probe_dec l p = 0 :=
  conj (fun ni : ~ In p l => match gt_0_eq (count_occ eq_Probe_dec l p) with
                               | or_introl gt0 => match ni (proj2 (count_occ_In _ _ _) gt0) with end
                               | or_intror eq0 => eq_sym eq0 end)
       (fun c0 => fun i => gt_irrefl _ (eq_ind_r _ (proj1 (count_occ_In _ _ _) i) c0)).

Definition count_sing_one {p} : count_occ eq_Probe_dec (p :: nil) p = 1 :=
  count_occ_cons_eq eq_Probe_dec _ eq_refl.

Fixpoint count_occ_app {l1 l2 p} : count_occ eq_Probe_dec (l1 ++ l2) p
                                 = count_occ eq_Probe_dec l1 p + count_occ eq_Probe_dec l2 p :=
  match l1 as l1 return
  count_occ eq_Probe_dec (l1 ++ l2) p
  = count_occ eq_Probe_dec l1 p + count_occ eq_Probe_dec l2 p with
    | nil => eq_ind_r (fun x => _ = x + _) (eq_sym (plus_O_n _)) (count_occ_nil _ _)
    | p' :: l1' => match eq_Probe_dec p' p with
                     | left e => eq_ind_r (fun x => _ = x + _)
                                          (eq_trans (eq_trans (count_occ_cons_eq eq_Probe_dec (l1' ++ l2) e)
                                                              (f_equal S (count_occ_app (l1 := l1'))))
                                                    (eq_sym (plus_Sn_m _ _)))
                                          (count_occ_cons_eq eq_Probe_dec l1' e)
                     | right ne => eq_ind_r (fun x => _ = x + _)
                                            (eq_trans (count_occ_cons_neq eq_Probe_dec (l1' ++ l2) ne)
                                                      (count_occ_app (l1 := l1')))
                                            (count_occ_cons_neq eq_Probe_dec l1' ne)
                   end
  end.

Definition count_occ_cons {l1 l2 p1 p2 n}
                          (c : count_occ eq_Probe_dec l1 p1 + n = count_occ eq_Probe_dec l2 p1)
                             : count_occ eq_Probe_dec (p2 :: l1) p1 + n = count_occ eq_Probe_dec (p2 :: l2) p1 :=
  match eq_Probe_dec p2 p1 with
    | right ne => eq_ind_r (fun x => _ = x)
                           (eq_ind_r (fun x => x + n = _) c (count_occ_cons_neq _ _ ne))
                           (count_occ_cons_neq _ _ ne)
    | left e => eq_trans (eq_ind_r (fun x => x + n = _) (eq_refl _) (count_occ_cons_eq _ _ e))
                         (eq_trans (f_equal S c) (eq_sym (count_occ_cons_eq _ _ e))) end.

Theorem count_occ_remove_neq : forall {p1 p2 l} (ne : p1 <> p2),
  count_occ eq_Probe_dec (remove eq_Probe_dec p2 l) p1 = count_occ eq_Probe_dec l p1.
intros. induction l; simpl.
  reflexivity.
  destruct (eq_Probe_dec p2 a).
    subst. destruct (eq_Probe_dec a p1).
      destruct (ne (eq_sym e)).
      assumption.
    simpl. destruct (eq_Probe_dec a p1).
      f_equal. assumption.
      assumption.
Qed.

Theorem count_occ_remove : forall {s p1 p2 n}
                                   (c1 : count_occ eq_Probe_dec (p2 :: nil) p1 + n = count_occ eq_Probe_dec s p1)
                                   (c2 : count_occ eq_Probe_dec s p2 = 0 \/ count_occ eq_Probe_dec s p2 = 1),
                                        count_occ eq_Probe_dec nil p1 + n
                                      = count_occ eq_Probe_dec (remove eq_Probe_dec p2 s) p1.
intros. simpl in *. destruct (eq_Probe_dec p2 p1).
  subst. simpl in c1. destruct c2.
    rewrite H in c1. inversion c1.
    rewrite H in c1. inversion c1. clear c1 H H1 n. induction s; simpl.
      reflexivity.
      destruct (eq_Probe_dec p1 a).
        assumption.
        simpl. destruct (eq_Probe_dec a p1).
          destruct (n (eq_sym e)).
          assumption.
  simpl in c1. subst. symmetry. apply count_occ_remove_neq. intro. exact (n0 (eq_sym H)).
Qed.

Fixpoint in_count_occ_app_one {p l1 l2} (f : count_occ eq_Probe_dec (l1 ++ l2) p = 1)
  : (In p l1 -> ~ In p l2 /\ count_occ eq_Probe_dec l1 p = 1)
 /\ (In p l2 -> ~ In p l1 /\ count_occ eq_Probe_dec l2 p = 1) :=
  match plus_is_one _ _ (eq_trans (eq_sym count_occ_app) f) with
    | left (conj L0 R1) => let L := proj2 count_occ_not_In L0
                           in conj (fun i => match L i with end)
                                   (fun i => conj L R1)
    | right (conj L1 R0) => let R := proj2 count_occ_not_In R0
                            in conj (fun i => conj R L1)
                                    (fun i => match R i with end)
  end.

Fixpoint in_once_or_more {p l} (i : In p l) : {count_occ eq_Probe_dec l p = 1}
                                            + {count_occ eq_Probe_dec l p > 1} :=
  match l as l return In p l -> {count_occ eq_Probe_dec l p = 1}
                              + {count_occ eq_Probe_dec l p > 1} with
    | nil => fun i => match in_nil i with end
    | x :: l' => fun i => match eq_Probe_dec x p, in_dec eq_Probe_dec p l' with
                            | left e, left i' =>
                              right (eq_ind_r (fun x => x > 1)
                                              (match in_once_or_more i' with
                                                 | left eq1 => eq_ind_r (fun x => S x > 1) (gt_Sn_n _) eq1
                                                 | right gt1 => gt_trans _ _ _ (gt_Sn_n _) gt1 end)
                                              (count_occ_cons_eq _ _ e))
                            | right ne, left i' =>
                              match in_once_or_more i' with
                                | left eq1 => left (eq_trans (count_occ_cons_neq _ _ ne) eq1)
                                | right gt1 => right (eq_ind_r (fun x => x > 1) gt1 (count_occ_cons_neq _ _ ne)) end
                            | left e, right ni' =>
                              left (eq_ind_r (fun x => _ = S x)
                                             (count_occ_cons_eq _ _ e)
                                             (eq_sym (proj1 count_occ_not_In ni')))
                            | right ne, right ni' => match match in_inv i with
                                                       | or_introl e => ne e
                                                       | or_intror i' => ni' i' end with end end
  end i.



(* ~~~~~~~ PROOF ABOUT LT ~~~~~~~ *)
Definition lt_not_eq {n m} (nltm : n < m) : n <> m :=
  fun e => gt_irrefl _ (eq_ind_r (fun x => _ < x) nltm e).



(* ~~~~~~~ PROOFS ABOUT PLUS ~~~~~~~ *)

Definition eq_plus_assoc {a b c n s} (e : c = a + b \/ c = b + a) : (c + n = s) <-> a + (b + n) = s :=
  (fun e => conj (fun x => eq_trans (plus_assoc _ _ _) (eq_ind_r (fun x => x + n = s) x (eq_sym e)))
                 (fun x => eq_ind_r (fun x => x + n = s) (eq_trans (eq_sym (plus_assoc _ _ _)) x) e))
  (match e return c = a + b with | or_introl e => e | or_intror e' => eq_trans e' (plus_comm _ _) end).

Definition plus_gt_one {n m} (nm : n > 1 \/ m > 1 \/ (n > 0 /\ m > 0)) : n + m > 1 :=
  match nm with
    | or_introl ngt1 => le_trans _ _ _ ngt1 (le_plus_l _ _)
    | or_intror (or_introl mgt1) => le_trans _ _ _ mgt1 (le_plus_r _ _)
    | or_intror (or_intror (conj ngt0 mgt0)) => plus_le_compat _ _ _ _ (gt_le_S _ _ ngt0) (gt_le_S _ _ mgt0)
  end.

Definition plus_not_le_one {n m} (ngt0 : n > 0) (mgt0 : m > 0) : ~ n + m <= 1 :=
  fun nmle1 => le_not_gt _ _ nmle1 (le_S_gt _ _ (plus_gt_one (or_intror (or_intror (conj ngt0 mgt0))))).

Definition plus_le_or_l {n m o} (nmleo : n + m <= o) (ogtn : o > n) : n + m < o \/ m > 0 :=
  match le_lt_or_eq _ _ nmleo with
    | or_introl lt => or_introl lt
    | or_intror eq => or_intror (plus_lt_reg_l _ _ _
                                (eq_ind_r (fun x => _ > x)
                                          (plus_lt_compat_r _ _ _ ogtn)
                                          (eq_sym (eq_trans eq (plus_n_O _)))))
  end.

Definition plus_le_or_r {n m o} (nmleo : n + m <= o) (ogtm : o > m) : n + m < o \/ n > 0 :=
  match plus_le_or_l (eq_ind_r (fun x => x <= _) nmleo (plus_comm _ _)) ogtm with
    | or_introl lt => or_introl (eq_ind_r (fun x => x < _) lt (plus_comm _ _))
    | or_intror gt => or_intror gt
  end.

Definition le_trans_plus {n m o} (nmleo : n + m <= o) : n <= o /\ m <= o :=
  conj (le_trans _ _ _ (le_plus_l _ _) nmleo)
       (le_trans _ _ _ (le_plus_r _ _) nmleo).

Theorem count_cons_eq : forall {l l' p p' n m}, l + n = p + m -> l' = S l -> S p = p' -> l' + n = p' + m.
intros. subst. change ((1 + l) + n = (1 + p) + m). rewrite <- plus_assoc. rewrite <- plus_assoc. f_equal. assumption.
Qed.

Theorem count_trans_neq : forall {l l' p p' n m}, l + n = p + m -> l' = l -> p = p' -> l' + n = p' + m.
intros. subst. assumption.
Qed.

Theorem count_remove_eq : forall {l l' p p' n m}, l + n = p + m -> l' = 0 -> l = 1 -> p = 1 -> p' = 0 -> l'+ n = p' + m.
intros. subst. inversion H. reflexivity.
Qed.

Theorem plus_swap : forall {x a b c}, x = a + (b + c) -> x = b + (a + c).
intros. subst. apply plus_permute.
Qed.