(* ProofStateToTraceBase contains definitions and proofs common to
   ProofStateToTraceCompliance, ProofStateToTraceInterferance,
   and ProofStateToTraceStability,
   including:
    - Program_to_NoopPre, Program_to_NoopReq
    - SS_to_Pre
    - es_interferant, es_nondetreq, es_nonstable, es_dec *)



Require Import List.
Require Import SemanticsBase.
Require Import SemanticsState.
Require Import SemanticsTrace.
Require Import ProofPre.
Require Import ProofBase.
Require Import ProofStateToTraceMap.



(* ~~~~~~~ NO-OP STEPS TO PRE, REQ ~~~~~~~ *)

Fixpoint Program_to_NoopPre (P : Program) : pre (Program_to_NoopLabel P) = Program_to_State P :=
  match P as P return pre (Program_to_NoopLabel P) = Program_to_State P with
    | comm (bang' A) => eq_refl _
    | comm (quer' A) => eq_refl _
    | sequ P0 P1 => Program_to_NoopPre P0
    | par P0 P1 => f_equal2 _ (Program_to_NoopPre P0) (Program_to_NoopPre P1)
    | _ => eq_refl _
  end.

Fixpoint Program_to_NoopReq {l} (P : Program) : req (Program_to_NoopLabel P) l :=
  match P as P return req (Program_to_NoopLabel P) l with
    | comm (bang' A) => I
    | comm (quer' A) => I
    | sequ P0 P1 => Program_to_NoopReq P0
    | par P0 P1 => conj (Program_to_NoopReq P0) (Program_to_NoopReq P1)
    | _ => I
  end.



(* ~~~~~~~ NON-ERRONEOUS STEPS TO PRE ~~~~~~~ *)

Fixpoint SS_to_Pre {P P' s s'} (ss : SmallStep P s P' s') : pre (SS_to_Label ss) = Program_to_State P :=
  match ss as ss in SmallStep P s P' s' return pre (SS_to_Label ss) = Program_to_State P with
    | Sequ_S _ _ _ _ _ ss' => SS_to_Pre ss'
    | ParL_S _ _ P2 _ _ ss' => f_equal2 _ (SS_to_Pre ss') (Program_to_NoopPre P2)
    | ParR_S P1 _ _ _ _ ss' => f_equal2 _ (Program_to_NoopPre P1) (SS_to_Pre ss')
    | _ => eq_refl _
  end.



(* ~~~~~~~ ERRONEOUS STEP ERROR TYPES ~~~~~~~ *)

Fixpoint es_interferant {P s} (es : ErrorStep P s) : Prop :=
  match es with
    | CommB2_E _ _ _ _ => True
    | CommQ2_E _ _ _ _ => True
    | Sequ_E _ _ _ es' => es_interferant es'
    | ParL_E _ _ _ es' => es_interferant es'
    | ParR_E _ _ _ es' => es_interferant es'
    | _ => False
  end.

Fixpoint es_nondetreq {P s} (es : ErrorStep P s) : Prop :=
  match es with
    | Sequ_E _ _ _ es' => es_nondetreq es'
    | ParL_E _ _ _ es' => es_nondetreq es'
    | ParR_E _ _ _ es' => es_nondetreq es'
    | Det2_E _ _ _ _ _ => True
    | _ => False
  end.

Fixpoint es_nonstable {P s} (es : ErrorStep P s) : Prop :=
  match es with
    | Sequ_E _ _ _ es' => es_nonstable es'
    | ParL_E _ _ _ es' => es_nonstable es'
    | ParR_E _ _ _ es' => es_nonstable es'
    | ParL_NSE _ _ _ _ _ _ _ => True
    | ParR_NSE _ _ _ _ _ _ _ => True
    | _ => False
  end.

Fixpoint es_dec {P s} (es : ErrorStep P s) : sumor (sumor (es_interferant es)
                                                          (es_nondetreq es))
                                                          (es_nonstable es) :=
  match es as es return sumor (sumor (es_interferant es)
                                     (es_nondetreq es))
                                     (es_nonstable es) with
    | CommB2_E _ _ _ _ => inleft (inleft I)
    | CommQ2_E _ _ _ _ => inleft (inleft I)
    | Sequ_E _ _ _ es' => es_dec es'
    | ParL_E _ _ _ es' => es_dec es'
    | ParR_E _ _ _ es' => es_dec es'
    | Det2_E _ _ _ _ _ => inleft (inright I)
    | ParL_NSE _ _ _ _ _ _ _ => inright I
    | ParR_NSE _ _ _ _ _ _ _ => inright I
  end.