(* ProofStateToTraceInterferance proves that
   SmallSteps map to noninterferant Traces
   and interferant ErrorSteps map to interferant Traces,
   defining:
    - SS_to_notInt
    - ES_to_Int *)



Require Import List.
Require Import SemanticsBase.
Require Import SemanticsState.
Require Import SemanticsTrace.
Require Import ProofPre.
Require Import ProofBase.
Require Import ProofStateToTraceMap.
Require Import ProofStateToTraceBase.



(* ~~~~~~~ NO-OP STEPS TO UP, DOWN ~~~~~~~ *)

Fixpoint Program_to_NoopUp (P : Program) : up (Program_to_NoopLabel P) = Program_to_State P :=
  match P as P return up (Program_to_NoopLabel P) = Program_to_State P with
    | comm (bang' A) => eq_refl _
    | comm (quer' A) => eq_refl _
    | sequ P0 P1 => Program_to_NoopUp P0
    | par P0 P1 => f_equal2 _ (Program_to_NoopUp P0) (Program_to_NoopUp P1)
    | _ => eq_refl _
  end.

Fixpoint Program_to_NoopDown {p} (P : Program) : ~ In p (down (Program_to_NoopLabel P)) :=
  match P as P return ~ In p (down (Program_to_NoopLabel P)) with
    | comm (bang' A) => in_nil (a:=p)
    | comm (quer' A) => in_nil (a:=p)
    | sequ P0 P1 => Program_to_NoopDown P0
    | par P0 P1 => fun i => match in_app_or _ _ p i with
                              | or_introl il => Program_to_NoopDown P0 il
                              | or_intror ir => Program_to_NoopDown P1 ir
                            end
    | _ => in_nil (a:=p)
  end.



(* ~~~~~~~ PROPER ~~~~~~~ *)

Definition proper (P : Program) (s : State) : Prop :=
  forall p, count_occ eq_Probe_dec (Program_to_State P) p
         <= count_occ eq_Probe_dec s p.

Definition exact_to_proper {P s} (e : exact P s) : proper P s :=
  fun p => eq_ind_r (fun x => x <= _) (le_n _) (e p).

Definition proper_par {P Q s} (p : proper (par P Q) s) : proper P s /\ proper Q s :=
  conj (fun p' => proj1 (le_trans_plus (eq_ind_r (fun x => x <= _) (p p') (eq_sym count_occ_app))))
       (fun p' => proj2 (le_trans_plus (eq_ind_r (fun x => x <= _) (p p') (eq_sym count_occ_app)))).



(* ~~~~~~~ NON-ERRONEOUS STEPS TO NON-INTERFERANCE ~~~~~~~ *)

Fixpoint SS_to_DownCount {P P' s s' p} (ss : SmallStep P s P' s') (id : In p (down (SS_to_Label ss)))
                                       : count_occ eq_Probe_dec s p = 1 :=
  match ss as ss in SmallStep P s P' s'
  return In p (down (SS_to_Label ss)) -> count_occ eq_Probe_dec s p = 1 with
    | CommB2_S A _ _ c2 => fun id => match eq_Probe_dec p (bar A) with
                                       | right pnbA => match not_in_neq pnbA id with end
                                       | left pebA => eq_ind_r (fun x => count_occ _ _ x = 1) c2 pebA end
    | CommQ2_S A _ _ c2 => fun id => match eq_Probe_dec p (hat A) with
                                       | right pnhA => match not_in_neq pnhA id with end
                                       | left pehA => eq_ind_r (fun x => count_occ _ _ x = 1) c2 pehA end
    | Sequ_S _ _ _ _ _ ss' => fun id => SS_to_DownCount ss' id
    | ParL_S _ _ _ _ _ ss' => fun id => match in_app_or _ _ _ id with
                                          | or_intror idR => match Program_to_NoopDown _ idR with end
                                          | or_introl idL => SS_to_DownCount ss' idL end
    | ParR_S _ _ _ _ _ ss' => fun id => match in_app_or _ _ _ id with
                                          | or_intror idR => SS_to_DownCount ss' idR
                                          | or_introl idL => match Program_to_NoopDown _ idL with end end
    | _ => fun id => match in_nil id with end
  end id.

Fixpoint SS_to_notInt' {P P' s s'} (ss : SmallStep P s P' s') (p : proper P s)
                                   : interferant (SS_to_Label ss) -> False :=
  match ss as ss in SmallStep P s P' s'
  return proper P s -> interferant (SS_to_Label ss) -> False with
    | CommB2_S A _ _ _ => fun _ => fun i => match i with existT p (conj piu pid) =>
                                            match eq_Probe_dec p (bar A) with
                                              | left pebA => in_nil piu
                                              | right pnbA => not_in_neq pnbA pid
                                            end end
    | CommQ2_S A _ _ _ => fun _ => fun i => match i with existT p (conj piu pid) =>
                                            match eq_Probe_dec p (hat A) with
                                              | left pebA => in_nil piu
                                              | right pnbA => not_in_neq pnbA pid
                                            end end
    | Sequ_S _ _ _ _ _ ss' => fun p => SS_to_notInt' ss' p
    | ParL_S _ _ _ _ _ ss' => fun p => fun i => match i with existT p' (conj piu pid) =>
                                                match in_app_or _ _ _ pid with
                                                  | or_intror pidR => Program_to_NoopDown _ pidR
                                                  | or_introl pidL =>
                                                match in_app_or _ _ _ piu with
                                                  | or_introl piuL =>
                                                    SS_to_notInt' ss' (proj1 (proper_par p)) (existT _ p' (conj piuL pidL))
                                                  | or_intror piuR =>
                                                    plus_not_le_one (proj1 (count_occ_In _ _ _ )
                                                                           (eq_ind_r (fun x => In _ x)
                                                                                     (down_to_pre pidL)
                                                                                     (eq_sym (SS_to_Pre ss'))))
                                                                    (proj1 (count_occ_In _ _ _)
                                                                           (eq_ind_r (fun x => In _ x) piuR
                                                                                     (eq_sym (Program_to_NoopUp _))))
                                                                    (eq_ind_r (fun x => _ <= x)
                                                                              (eq_ind_r (fun x => x <= _) (p p') (eq_sym count_occ_app))
                                                                              (eq_sym (SS_to_DownCount ss' pidL))) end end end
    | ParR_S p _ _ _ _ ss' => fun p => fun i => match i with existT p' (conj piu pid) =>
                                                match in_app_or _ _ _ pid with
                                                  | or_introl pidL => Program_to_NoopDown _ pidL
                                                  | or_intror pidR =>
                                                match in_app_or _ _ _ piu with
                                                  | or_intror piuR =>
                                                    SS_to_notInt' ss' (proj2 (proper_par p)) (existT _ p' (conj piuR pidR))
                                                  | or_introl piuL =>
                                                    plus_not_le_one (proj1 (count_occ_In _ _ _)
                                                                           (eq_ind_r (fun x => In _ x) piuL
                                                                                     (eq_sym (Program_to_NoopUp _))))
                                                                    (proj1 (count_occ_In _ _ _ )
                                                                           (eq_ind_r (fun x => In _ x)
                                                                                     (down_to_pre pidR)
                                                                                     (eq_sym (SS_to_Pre ss'))))
                                                                    (eq_ind_r (fun x => _ <= x)
                                                                              (eq_ind_r (fun x => x <= _) (p p') (eq_sym count_occ_app))
                                                                              (eq_sym (SS_to_DownCount ss' pidR))) end end end
    | _ => fun _ => fun i => match i with existT _ piud => match piud with conj _ pid => in_nil pid end end
  end p.

Definition SS_to_notInt {P P' s s'} (ss : SmallStep P s P' s') (e : exact P s)
                                    : interferant (SS_to_Label ss) -> False :=
  SS_to_notInt' ss (exact_to_proper e).



(* ~~~~~~~ ERRONEOUS STEPS TO INTERFERANCE ~~~~~~~ *)

Fixpoint ES_to_Int' {P s} (es : ErrorStep P s)
                           (p : proper P s) (i : es_interferant es)
                              : {p : Probe & In p (down (ES_to_Label es))
                                           & In p (up (ES_to_Label es))
                                          \/ count_occ eq_Probe_dec s p
                                           > count_occ eq_Probe_dec (Program_to_State P) p} :=
  match es as es in ErrorStep P s
  return proper P s -> es_interferant es
  -> {p : Probe & In p (down (ES_to_Label es))
                & In p (up (ES_to_Label es))
               \/ count_occ eq_Probe_dec s p
                > count_occ eq_Probe_dec (Program_to_State P) p} with
    | CommB2_E A _ _ c2 => fun _ => fun _ => existT2 (fun x => In x _) _ (bar A) (in_eq _ _)
                                                              (or_intror (eq_ind_r (fun x => _ > x) c2 count_sing_one))
    | CommQ2_E A _ _ c2 => fun _ => fun _ => existT2 (fun x => In x _) _ (hat A) (in_eq _ _)
                                                              (or_intror (eq_ind_r (fun x => _ > x) c2 count_sing_one))
    | Sequ_E _ _ _ es' => fun p => fun i => ES_to_Int' es' p i
    | ParL_E _ _ _ es' => fun p => fun i => match ES_to_Int' es' (proj1 (proper_par p)) i with existT2 p' id iuc =>
                                                     existT2 _ _ p' (in_or_app _ _ _ (or_introl id))
                                                     match iuc return _ \/ _ with
                                                       | or_introl iu => or_introl (in_or_app _ _ _ (or_introl iu))
                                                       | or_intror c =>
                                                     match plus_le_or_l (eq_ind_r (fun x => x <= _) (p p') (eq_sym count_occ_app)) c with
                                                       | or_introl lts => or_intror (eq_ind_r (fun x => x < _) lts count_occ_app)
                                                       | or_intror gt0 => or_introl (in_or_app _ _ _ (or_intror
                                                                                    (eq_ind_r (fun x => In _ x)
                                                                                              (proj2 (count_occ_In _ _ _) gt0)
                                                                                              (Program_to_NoopUp _)))) end end end
    | ParR_E _ _ _ es' => fun p => fun i => match ES_to_Int' es' (proj2 (proper_par p)) i with existT2 p' id iuc =>
                                                     existT2 _ _ p' (in_or_app _ _ _ (or_intror id))
                                                     match iuc return _ \/ _ with
                                                       | or_introl iu => or_introl (in_or_app _ _ _ (or_intror iu))
                                                       | or_intror c =>
                                                     match plus_le_or_r (eq_ind_r (fun x => x <= _) (p p') (eq_sym count_occ_app)) c with
                                                       | or_introl lts => or_intror (eq_ind_r (fun x => x < _) lts count_occ_app)
                                                       | or_intror gt0 => or_introl (in_or_app _ _ _ (or_introl
                                                                                    (eq_ind_r (fun x => In _ x)
                                                                                              (proj2 (count_occ_In _ _ _) gt0)
                                                                                              (Program_to_NoopUp _)))) end end end
    | _ => fun _ => fun i => match i with end
  end p i.

Definition ES_to_Int {P s} (es : ErrorStep P s)
                            (e : exact P s) (i : es_interferant es)
                               : interferant (ES_to_Label es) :=
  match ES_to_Int' es (exact_to_proper e) i with existT2 p id iuc =>
  existT _ p (conj (match iuc with
                      | or_introl iu => iu
                      | or_intror c' => match lt_not_eq c' (e p) with end end) id) end.