(* ProofStateToTraceStability proves that
   nonstable ErrorSteps map to nonstable Traces,
   defining:
    - ES_to_notStable *)



Require Import List.
Require Import SemanticsBase.
Require Import SemanticsState.
Require Import SemanticsTrace.
Require Import ProofPre.
Require Import ProofBase.
Require Import ProofStateToTraceMap.
Require Import ProofStateToTraceBase.



(* ~~~~~~~ NO-OP STEPS TO STATE, POST ~~~~~~~ *)

Fixpoint Program_to_NoopLabelState {s} (P : Program) : Label_to_newState (Program_to_NoopLabel P) s = s :=
  match P as P return Label_to_newState (Program_to_NoopLabel P) s = s with
    | comm (bang' A) => eq_refl _
    | comm (quer' A) => eq_refl _
    | sequ P0 P1 => Program_to_NoopLabelState P0
    | par P0 P1 => eq_trans (Program_to_NoopLabelState P1) (Program_to_NoopLabelState P0)
    | _ => eq_refl _
  end.

Fixpoint Program_to_NoopPost (P : Program) : Program_to_State P = post (Program_to_NoopLabel P) :=
  match P as P return Program_to_State P = post (Program_to_NoopLabel P) with
    | comm (bang' A) => eq_refl _
    | comm (quer' A) => eq_refl _
    | sequ P0 P1 => Program_to_NoopPost P0
    | par P0 P1 => f_equal2 _ (Program_to_NoopPost P0) (Program_to_NoopPost P1)
    | _ => eq_refl _
  end.



(* ~~~~~~~ NON-ERRONEOUS STEPS TO STATE, POST, COUNT ~~~~~~~ *)

Fixpoint SS_to_LabelState {P P' s s'} (ss : SmallStep P s P' s') : Label_to_newState (SS_to_Label ss) s = s' :=
  match ss as ss in SmallStep P s P' s' return Label_to_newState (SS_to_Label ss) s = s' with
    | Sequ_S _ _ _ _ _ ss' => SS_to_LabelState ss'
    | ParL_S _ _ P2 _ _ ss' => eq_trans (Program_to_NoopLabelState P2) (SS_to_LabelState ss')
    | ParR_S P1 _ _ _ _ ss' => eq_ind_r (fun x => Label_to_newState _ x = _)
                                        (SS_to_LabelState ss')
                                        (Program_to_NoopLabelState P1)
    | _ => eq_refl _
  end.

Fixpoint SS_to_Post {P P' s s'} (ss : SmallStep P s P' s') (v : valid P)
                    : Program_to_State P' = post (SS_to_Label ss) :=
  match ss as ss in SmallStep P s P' s'
  return valid P -> Program_to_State P' = post (SS_to_Label ss) with
    | Sequ_S _ _ _ _ _ ss' => fun v => SS_to_Post ss' (proj1 v)
    | SequSkip_S _ _ => fun v => initial_to_nil (proj2 v)
    | ParL_S _ _ P2 _ _ ss' => fun v => f_equal2 _ (SS_to_Post ss' (proj1 v)) (Program_to_NoopPost P2)
    | ParR_S P1 _ _ _ _ ss' => fun v => f_equal2 _ (Program_to_NoopPost P1) (SS_to_Post ss' (proj2 v))
    | Rep_S P _ => fun v => initial_to_nil v
    | DetL_S _ _ _ _ => fun v => initial_to_nil (proj1 v)
    | DetR_S _ _ _ _ => fun v => initial_to_nil (proj2 v)
    | NDetL_S _ _ _ _ => fun v => initial_to_nil (proj1 v)
    | NDetR_S _ _ _ _ => fun v => initial_to_nil (proj2 v)
    | _ => fun _ => eq_refl _
  end v.

Fixpoint SS_Count {P P' s s' p n} (ss : SmallStep P s P' s') (v : valid P)
                  (c : count_occ eq_Probe_dec (Program_to_State P) p + n =
                       count_occ eq_Probe_dec s p)
                  : count_occ eq_Probe_dec (Program_to_State P') p + n =
                    count_occ eq_Probe_dec s' p :=
  match ss in SmallStep P s P' s' return valid P
  -> count_occ eq_Probe_dec (Program_to_State P) p + n = count_occ eq_Probe_dec s p
  -> count_occ eq_Probe_dec (Program_to_State P') p + n = count_occ eq_Probe_dec s' p with
    | CommB1_S _ _ _ => fun _ => fun c => count_occ_cons c
    | CommB2_S _ _ _ c2 => fun _ => fun c => count_occ_remove c (or_intror c2)
    | CommQ1_S _ _ _ => fun _ => fun c => count_occ_cons c
    | CommQ2_S _ _ _ c2 => fun _ => fun c => count_occ_remove c (or_intror c2)
    | Sequ_S _ _ _ _ _ ss' => fun v => fun c => SS_Count ss' (proj1 v) c
    | SequSkip_S _ _ => fun v => fun c => eq_ind_r (fun x => count_occ _ x _ + _ = _) c
                                                   (initial_to_nil (proj2 v))
    | ParL_S _ _ _ _ _ ss' => fun v => fun c => proj2 (eq_plus_assoc (or_introl count_occ_app))
                                                      (SS_Count ss' (proj1 v)
                                                      (proj1 (eq_plus_assoc (or_introl count_occ_app)) c))
    | ParR_S _ _ _ _ _ ss' => fun v => fun c => proj2 (eq_plus_assoc (or_intror count_occ_app))
                                                      (SS_Count ss' (proj2 v)
                                                      (proj1 (eq_plus_assoc (or_intror count_occ_app)) c))
    | ParSkip_S _ => fun _ => fun c => c
    | Rep_S _ _ => fun v => fun c => eq_ind_r (fun x => count_occ _ x _ + _ = _) c
                                              (initial_to_nil v)
    | DetL_S _ _ _ _ => fun v => fun c => eq_ind_r (fun x => count_occ _ x _ + _ = _) c
                                                   (initial_to_nil (proj1 v))
    | DetR_S _ _ _ _ => fun v => fun c => eq_ind_r (fun x => count_occ _ x _ + _ = _) c
                                                   (initial_to_nil (proj2 v))
    | NDetL_S _ _ _ _ => fun v => fun c => eq_ind_r (fun x => count_occ _ x _ + _ = _) c
                                                    (initial_to_nil (proj1 v))
    | NDetR_S _ _ _ _ => fun v => fun c => eq_ind_r (fun x => count_occ _ x _ + _ = _) c
                                                    (initial_to_nil (proj2 v))
  end v c.



(* ~~~~~~~ ERRONEOUS STEPS TO NON-STABILITY ~~~~~~~ *)

Fixpoint NS_to_Post {P s s'} (ns : Nonstable P s s') (v : valid P)
                    : Program_to_State (NS_to_Program ns) = post (NS_to_Label ns) :=
  match ns as ns in Nonstable P s s'
  return valid P -> Program_to_State (NS_to_Program ns) = post (NS_to_Label ns) with
    | Sequ_NS _ _ _ _ ns' => fun v => NS_to_Post ns' (proj1 v)
    | ParL_NS _ P2 _ _ ns' => fun v => f_equal2 _ (NS_to_Post ns' (proj1 v)) (Program_to_NoopPost P2)
    | ParR_NS P1 _ _ _ ns' => fun v => f_equal2 _ (Program_to_NoopPost P1) (NS_to_Post ns' (proj2 v))
    | DetL_NS GP0 GP1 _ _ _ _ => fun v => initial_to_nil (proj1 v)
    | DetR_NS GP0 GP1 _ _ _ _ => fun v => initial_to_nil (proj2 v)
    | NDetL_NS GP0 GP1 _ _ _ _ => fun v => initial_to_nil (proj1 v)
    | NDetR_NS GP0 GP1 _ _ _ _ => fun v => initial_to_nil (proj2 v)
    | _ => fun _ => eq_refl _
  end v.

Fixpoint NS_to_notReq {P s s'} (ns : Nonstable P s s') : ~ req (NS_to_Label ns) s' :=
  match ns as ns in Nonstable P s s' return ~ req (NS_to_Label ns) s' with
    | CommB1_NS _ _ _ _ c2 => fun r => r c2
    | CommB2_NS _ _ _ _ c2 => c2
    | CommQ1_NS _ _ _ _ c2 => c2
    | CommQ2_NS _ _ _ _ c2 => fun r => r c2
    | Sequ_NS _ _ _ _ ns' => NS_to_notReq ns'
    | ParL_NS _ _ _ _ ns' => fun r => match r with conj rl _ => NS_to_notReq ns' rl end
    | ParR_NS _ _ _ _ ns' => fun r => match r with conj _ rr => NS_to_notReq ns' rr end
    | DetL_NS _ _ _ _ _ c2 => fun r => c2 (proj1 (eval_s1 _) r)
    | DetR_NS _ _ _ _ _ c2 => fun r => c2 (proj1 (eval_s1 _) r)
    | NDetL_NS _ _ _ _ _ c2 => fun r => c2 (proj1 (eval_s1 _) r)
    | NDetR_NS _ _ _ _ _ c2 => fun r => c2 (proj1 (eval_s1 _) r)
  end.

Fixpoint NS_Req {P s s' l} (ns : Nonstable P s s')
                            (r : req (NS_to_Label ns) (Label_to_newState (NS_to_Label ns) l))
                               : req (NS_to_Label ns) l :=
  match ns as ns in Nonstable P s s'
  return req (NS_to_Label ns) (Label_to_newState (NS_to_Label ns) l)
  -> req (NS_to_Label ns) l with
    | CommB1_NS A _ _ _ _ => fun r => fun i => r (in_cons _ _ _ i)
    | CommB2_NS A _ _ _ _ => fun r => in_remove_weak r
    | CommQ1_NS A _ _ _ _ => fun r => match in_inv r with
                                           | or_introl e => match proj2 neq_Probe e with end
                                           | or_intror r => r end
    | CommQ2_NS A _ _ _ _ => fun r => fun i => r (in_remove_neq (proj1 neq_Probe) i)
    | Sequ_NS _ _ _ _ ns' => fun r => NS_Req ns' r
    | ParL_NS _ P2 _ _ ns' => fun r => conj (NS_Req ns' (eq_ind_r (fun x => req _ x)
                                                                  (proj1 r)
                                                                  (eq_sym (Program_to_NoopLabelState P2))))
                                            (Program_to_NoopReq P2)
    | ParR_NS P1 _ _ _ ns' => fun r => conj (Program_to_NoopReq P1)
                                            (eq_ind_r (fun x => req _ x)
                                                      (NS_Req ns' (proj2 r))
                                                      (eq_sym (Program_to_NoopLabelState P1)))
    | _ => fun r => r
  end r.

Fixpoint NSE_inv {P P' s s' A} (ss : SmallStep P s P' s')
                 : prod (((~ In (bar A) s) /\ In (bar A) s') -> ~ In (hat A) s')
                        ((In (hat A) s /\ (~ In (hat A) s')) -> ~ In (bar A) s') :=
  match ss in SmallStep P s P' s'
  return prod (((~ In (bar A) s) /\ In (bar A) s') -> ~ In (hat A) s')
              ((In (hat A) s /\ (~ In (hat A) s')) -> ~ In (bar A) s') with
    | CommB1_S B _ c => match eq_Channel_dec A B with
                            | left e => (fun _ => fun i' => match in_inv i' with
                                                              | or_introl e' => proj1 neq_Probe e'
                                                              | or_intror i => eq_ind_r (fun x => ~ In x _) c (f_equal hat e) i end,
                                         fun ini => fun i => eq_ind_r (fun x => ~ In x _) c (f_equal hat e) (proj1 ini))
                            | right ne => (fun nii : ((~ In (bar A) _) /\ In (bar A) _) =>
                                           fun i' => match in_inv (proj2 nii) with
                                                       | or_introl e => proj1 (neq_Channel_to_Probe ne) (eq_sym e)
                                                       | or_intror i => proj1 nii i end,
                                           fun ini : (In (hat A) _ /\ (~ In (hat A) _)) =>
                                           fun i' => proj2 ini (in_cons _ _ _ (proj1 ini))) end
    | CommB2_S _ _ _ _ => (fun nii : ((~ In (bar A) _) /\ In (bar A) _) =>
                           fun i' => proj1 nii (in_remove_weak (proj2 nii)),
                           fun ini : (In (hat A) _ /\ (~ In (hat A) _)) =>
                           fun i' => proj2 ini (in_remove_neq (proj2 neq_Probe) (proj1 ini)))
    | CommQ1_S _ _ _ => (fun nii : ((~ In (bar A) _) /\ In (bar A) _) =>
                         fun i' => match in_inv (proj2 nii) with
                                     | or_introl e => (proj1 neq_Probe) (eq_sym e)
                                     | or_intror i => proj1 nii i end,
                         fun ini : (In (hat A) _ /\ (~ In (hat A) _)) =>
                         fun i' => proj2 ini (in_cons _ _ _ (proj1 ini)))
    | CommQ2_S B _ c1 _ => match eq_Channel_dec A B with
                            | left e => (fun nii : ((~ In (bar A) _) /\ In (bar A) _) =>
                                         fun i' => proj1 nii (in_remove_weak (proj2 nii)),
                                         fun _ => fun i' => eq_ind_r (fun x => ~ In x _) c1 (f_equal bar e) (in_remove_weak i'))
                            | right ne => (fun nii : ((~ In (bar A) _) /\ In (bar A) _) =>
                                           fun i' => proj1 nii (in_remove_weak (proj2 nii)),
                                           fun ini : (In (hat A) _ /\ (~ In (hat A) _)) =>
                                           fun i' => proj2 ini (in_remove_neq (proj2 (neq_Channel_to_Probe ne)) (proj1 ini))) end
    | Sequ_S _ _ _ _ _ ss' => NSE_inv ss'
    | ParL_S _ _ _ _ _ ss' => NSE_inv ss'
    | ParR_S _ _ _ _ _ ss' => NSE_inv ss'
    | _ => (fun nii : ((~ In (bar A) _) /\ In (bar A) _) =>
            fun i' => proj1 nii (proj2 nii),
            fun ini : (In (hat A) _ /\ (~ In (hat A) _)) =>
            fun i' => proj2 ini (proj1 ini))
  end.

Fixpoint NS_Count {Q P P' s s' p n} (ns : Nonstable Q s s') (ss : SmallStep P s P' s') (v : valid Q)
                  (c : count_occ eq_Probe_dec (Program_to_State Q) p + n = count_occ eq_Probe_dec s' p)
                     : count_occ eq_Probe_dec (Program_to_State (NS_to_Program ns)) p + n =
                       count_occ eq_Probe_dec (Label_to_newState (NS_to_Label ns) s') p :=
  match ns as ns in Nonstable Q s s' return SmallStep P s P' s' -> valid Q
  -> count_occ eq_Probe_dec (Program_to_State Q) p + n = count_occ eq_Probe_dec s' p
  -> count_occ eq_Probe_dec (Program_to_State (NS_to_Program ns)) p + n
             = count_occ eq_Probe_dec (Label_to_newState (NS_to_Label ns) s') p with
    | CommB1_NS _ _ _ _ _ => fun _ => fun _ => fun c => count_occ_cons c
    | CommB2_NS _ _ _ a b => fun ss => fun _ => fun c =>
                             count_occ_remove c (or_introl (proj1 count_occ_not_In
                                                (snd (NSE_inv ss) (conj a b))))
    | CommQ1_NS _ _ _ _ _ => fun _ => fun _ => fun c => count_occ_cons c
    | CommQ2_NS _ _ _ a b => fun ss => fun _ => fun c =>
                             count_occ_remove c (or_introl (proj1 count_occ_not_In
                                                (fst (NSE_inv ss) (conj a b))))
    | Sequ_NS _ _ _ _ ns' => fun ss => fun v => fun c => NS_Count ns' ss (proj1 v) c
    | ParL_NS _ q _ _ ns' => fun ss => fun v => fun c =>
                             eq_ind_r (fun x => _ = count_occ _ x p)
                                      (proj2 (eq_plus_assoc (or_introl count_occ_app))
                                      (NS_Count ns' ss (proj1 v)
                                      (proj1 (eq_plus_assoc (or_introl count_occ_app)) c)))
                                      (Program_to_NoopLabelState q)
    | ParR_NS q _ _ _ ns' => fun ss => fun v => fun c =>
                             eq_ind_r (fun x => _ = count_occ _ (Label_to_newState _ x) p)
                                      (proj2 (eq_plus_assoc (or_intror count_occ_app))
                                      (NS_Count ns' ss (proj2 v)
                                      (proj1 (eq_plus_assoc (or_intror count_occ_app)) c)))
                                      (Program_to_NoopLabelState q)
    | DetL_NS _ _ _ _ _ _ => fun ss => fun v => fun c =>
                             eq_ind_r (fun x => count_occ _ x _ + _ = _) c
                                      (initial_to_nil (proj1 v))
    | DetR_NS _ _ _ _ _ _ => fun ss => fun v => fun c =>
                             eq_ind_r (fun x => count_occ _ x _ + _ = _) c
                                      (initial_to_nil (proj2 v))
    | NDetL_NS _ _ _ _ _ _ => fun ss => fun v => fun c =>
                             eq_ind_r (fun x => count_occ _ x _ + _ = _) c
                                      (initial_to_nil (proj1 v))
    | NDetR_NS _ _ _ _ _ _ => fun ss => fun v => fun c =>
                             eq_ind_r (fun x => count_occ _ x _ + _ = _) c
                                      (initial_to_nil (proj2 v))
  end ss v c.

Fixpoint ES_to_State {P s} (es : ErrorStep P s) (ns : es_nonstable es) : State :=
  match es as es return es_nonstable es -> State with
    | Sequ_E _ _ _ es' => fun ns => ES_to_State es' ns
    | ParL_E _ P2 _ es' => fun ns => ES_to_State es' ns
    | ParR_E P1 _ _ es' => fun ns => ES_to_State es' ns
    | ParL_NSE _ _ _ _ s' _ ns => fun _ => Label_to_newState (NS_to_Label ns) s'
    | ParR_NSE _ _ _ _ s' _ ns => fun _ => Label_to_newState (NS_to_Label ns) s'
    | _ => fun ns => match ns with end
  end ns.

Fixpoint ES_to_Post {P s} (es : ErrorStep P s) (v : valid P)
                    : Program_to_State (ES_to_Program es) = post (ES_to_Label es) :=
  match es as es in ErrorStep P s
  return valid P -> Program_to_State (ES_to_Program es) = post (ES_to_Label es) with
    | Sequ_E _ _ _ es' => fun v => ES_to_Post es' (proj1 v)
    | ParL_E _ P2 _ es' => fun v => f_equal2 _ (ES_to_Post es' (proj1 v)) (Program_to_NoopPost P2)
    | ParR_E P1 _ _ es' => fun v => f_equal2 _ (Program_to_NoopPost P1) (ES_to_Post es' (proj2 v))
    | ParL_NSE _ _ _ _ _ ss ns => fun v => f_equal2 _ (SS_to_Post ss (proj1 v)) (NS_to_Post ns (proj2 v))
    | ParR_NSE _ _ _ _ _ ss ns => fun v => f_equal2 _ (NS_to_Post ns (proj1 v)) (SS_to_Post ss (proj2 v))
    | _ => fun _ => eq_refl _
  end v.

Fixpoint ES_to_notReq {P s} (es : ErrorStep P s) (ns : es_nonstable es)
                      : ~ req (ES_to_Label es) (ES_to_State es ns) :=
  match es as es return forall ns : es_nonstable es, ~ req (ES_to_Label es) (ES_to_State es ns) with
    | Sequ_E _ _ _ es' => fun ns => ES_to_notReq es' ns
    | ParL_E _ _ _ es' => fun ns => fun r => ES_to_notReq es' ns (proj1 r)
    | ParR_E _ _ _ es' => fun ns => fun r => ES_to_notReq es' ns (proj2 r)
    | ParL_NSE _ _ _ _ _ ss ns => fun _ => fun r => NS_to_notReq ns (NS_Req ns (proj2 r))
    | ParR_NSE _ _ _ _ _ ss ns => fun _ => fun r => NS_to_notReq ns (NS_Req ns (proj1 r))
    | _ => fun ns => match ns with end
  end ns.

Fixpoint ES_Count {P s p n} (es : ErrorStep P s) (ns : es_nonstable es) (v : valid P)
                             (c : count_occ eq_Probe_dec (Program_to_State P) p + n =
                                  count_occ eq_Probe_dec s p)
                                : count_occ eq_Probe_dec (Program_to_State (ES_to_Program es)) p + n =
                                  count_occ eq_Probe_dec (ES_to_State es ns) p :=
  match es as es in ErrorStep P s
  return forall ns : es_nonstable es, valid P
      -> count_occ eq_Probe_dec (Program_to_State P) p + n = count_occ eq_Probe_dec s p
      -> count_occ eq_Probe_dec (Program_to_State (ES_to_Program es)) p + n =
         count_occ eq_Probe_dec (ES_to_State es ns) p with
    | Sequ_E _ _ _ es' => fun ns => fun v => fun c => ES_Count es' ns (proj1 v) c
    | ParL_E _ _ _ es' => fun ns => fun v => fun c => proj2 (eq_plus_assoc (or_introl count_occ_app))
                                                            (ES_Count es' ns (proj1 v)
                                                            (proj1 (eq_plus_assoc (or_introl count_occ_app)) c))
    | ParR_E _ _ _ es' => fun ns => fun v => fun c => proj2 (eq_plus_assoc (or_intror count_occ_app))
                                                            (ES_Count es' ns (proj2 v)
                                                            (proj1 (eq_plus_assoc (or_intror count_occ_app)) c))
    | ParL_NSE _ q _ _ _ ss ns => fun _ => fun v => fun c =>
                                  (proj2 (eq_plus_assoc (or_intror count_occ_app))
                                  (NS_Count ns ss (proj2 v)
                                  (proj1 (eq_plus_assoc (or_intror count_occ_app))
                                  (proj2 (eq_plus_assoc (or_introl count_occ_app))
                                  (SS_Count ss (proj1 v)
                                  (proj1 (eq_plus_assoc (or_introl count_occ_app)) c))))))
    | ParR_NSE _ _ _ _ _ ss ns => fun _ => fun v => fun c =>
                                  (proj2 (eq_plus_assoc (or_introl count_occ_app))
                                  (NS_Count ns ss (proj1 v)
                                  (proj1 (eq_plus_assoc (or_introl count_occ_app))
                                  (proj2 (eq_plus_assoc (or_intror count_occ_app))
                                  (SS_Count ss (proj2 v)
                                  (proj1 (eq_plus_assoc (or_intror count_occ_app)) c))))))
    | _ => fun ns => match ns with end
  end ns v c.

Definition ES_to_notStable {P s} (es : ErrorStep P s) (ns : es_nonstable es)
                                 (v : valid P) (e : exact P s) : ~ stable (ES_to_Label es) :=
  fun s => ES_to_notReq es ns (req_eq_set (eq_ind_r (fun x => req _ x) s (ES_to_Post es v))
                              (eq_State_to_eq_set
                              (fun p => eq_trans (plus_n_O _)
                              (ES_Count es ns v (eq_trans (eq_sym (plus_n_O _)) (e p)))))).